/*
oinst - software installation management tool
Copyright (C) 2000, 2001, 2002 Ed L. Cashin

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <limits.h>
#include <assert.h>
#include <errno.h>
#include "elcwft.h"
#include "elcwft_p.h"
#include "xstrdup.h"
#include "xstradd.h"

void usage(void)
{
    fprintf(stderr, "usage: %s {pkgname} {instdir}\n", PROGNAME);
}

/* does pname match the name of this directory? */
int verify_cwd_pkg(const char *pname)
{
    char *buf, *p;
    int n;

    errno = 0;
    if ( (n = pathconf(".", _PC_PATH_MAX)) == -1) {
      if (errno == 0) {
	n = 2048;
      } else {
	perror("pathconf");
	exit(EXIT_FAILURE);
      }
    }
    printf("debug (%s): allocating %d-size buffer\n", __FUNCTION__, n);
    if (! (buf = malloc(sizeof(char) * n)) ) {
      perror("malloc");
      exit(EXIT_FAILURE);
    }
    if (! getcwd(buf, n)) {
      perror("getcwd");
      exit(EXIT_FAILURE);
    }
    if (! (p = strrchr(buf, '/')) ) {
      fprintf(stderr, "couldn't find slash in cwd: %s\n", buf);
      exit(EXIT_FAILURE);
    }
    ++p;

    printf("debug (%s): comparing \"%s\" and \"%s\"\n", __FUNCTION__, pname, p);
    n = ! strcmp(pname, p);
    
    printf("debug (%s): freeing buffer\n", __FUNCTION__);
    free(buf);
    return n;
}

static const char *dirname_sans_tdot(const char *dirname)
{
    char *p;
    int len;

    assert(dirname); 
    len = strlen(dirname);
    assert(len > 0);
    if (dirname[len - 1] == '.' && dirname[len - 2] == '/') {
      p = xstrdup(dirname);
      p[len - 2] = '\0';
    } else {
      p = (char *) dirname;
    }
    return p;
}

wft_ret_t print_fname(const char *fname, const struct stat *sb, void *data)
{
    (void) sb;
    (void) data;
    puts(fname);
    
    return WFT_PROCEED;
}

wft_ret_t ro_file(const char *fname, const struct stat *sb, void *data)
{
    (void) data;

    if (S_ISREG(sb->st_mode)) {
      printf("%35s ... ", fname);
      fflush(stdout);
      if (chmod(fname, sb->st_mode & ~(S_IWUSR | S_IWGRP | S_IWOTH)) == -1) {
	fprintf(stderr, "%s: could not chmod file (%s): %s\n",
		PROGNAME, fname, strerror(errno));
	exit(EXIT_FAILURE);
      }
      puts("ro");
    }    

    return WFT_PROCEED;
}

int dir_exists(const char *fname)
{
    struct stat buf;

    if (stat(fname, &buf) == -1) {
      switch (errno) {
	case ENOENT:
	case ELOOP:
	case EFAULT:
	case ENAMETOOLONG:
	  return 0;
	  break;		/* not reached */
	default:
	  break;
      }
      perror("stat");
      exit(EXIT_FAILURE);
    }

    return S_ISDIR(buf.st_mode);
}

wft_ret_t ensure_dir(const char *fname, const struct stat *sb, void *data)
{
    char *instdir;
    const char *fname_pp;	/* pretty-printing fname */
    char *pcheck;
    struct stat sbuf;

    if (! S_ISDIR(sb->st_mode))
      return WFT_PROCEED;
    
    instdir = (char *) data;
    fname_pp = dirname_sans_tdot(fname);
    pcheck = xstradd(instdir, "/", fname_pp, NULL);
    printf("%35s ... ", pcheck);
    fflush(stdout);
    if (stat(pcheck, &sbuf) == -1) {
      if (errno != ENOENT) {
	perror("stat");		/* todo: should mkdir on ENOENT */
	exit(EXIT_FAILURE);
      }
      if (mkdir(pcheck, 0777) == -1) {
	perror("mkdir");
	exit(EXIT_FAILURE);
      } else {
	puts("mkdir");
      }
    } else {
      puts("exists");
    }
    free(pcheck);
    if (fname_pp != fname)
      free((char *) fname_pp);

    return WFT_PROCEED;
}

wft_ret_t create_hardlink(const char *fname, const struct stat *sb, void *data)
{
    char *instdir;
    char *instlink;
    struct stat sbuf;

    if (! S_ISREG(sb->st_mode))
      return WFT_PROCEED;
    
    instdir = (char *) data;
    instlink = xstradd(instdir, "/", fname, NULL);
    printf("%35s ... ", instlink);
    fflush(stdout);
    if (stat(instlink, &sbuf) == -1) {
      if (errno != ENOENT) {
	perror("stat");		/* todo: should mkdir on ENOENT */
	exit(EXIT_FAILURE);
      }
      if (link(fname, instlink) == -1) {
	perror("link");
	exit(EXIT_FAILURE);
      } else {
	puts("linked");
      }
    } else {
      /* something's there.  the same file? */
      if (sbuf.st_ino == sb->st_ino)
	puts("link exists");
      else
	/* todo: backup and replace existing files with different inode */
	puts("other file exists");
    }
    free(instlink);

    return WFT_PROCEED;
}

wft_ret_t create_symlink(const char *fname, const struct stat *sb, void *data)
{
    char *instdir;
    char *instlink;
    struct stat sbuf;

    if (! S_ISLNK(sb->st_mode))
      return WFT_PROCEED;
    
    instdir = (char *) data;
    instlink = xstradd(instdir, "/", fname, NULL);
    printf("%35s ... ", instlink);
    fflush(stdout);
    if (lstat(instlink, &sbuf) == -1) {
      if (errno != ENOENT) {
	perror("lstat");
	exit(EXIT_FAILURE);
      }
      if (symlink(instlink, fname) == -1) { 
	perror("link");
	/* exit(EXIT_FAILURE); */
      } else {
	puts("linked");
      }
    } else {
      /* something's there.  the same file? */
      if (sbuf.st_ino == sb->st_ino)
	puts("link exists");
      else
	/* todo: backup and replace existing files with different inode */
	puts("other file exists");
    }
    free(instlink);

    return WFT_PROCEED;
}

void init_wftctx(wft_context_t *w)
{
    w->rootname = ".";
    w->callback = ro_file;
    w->cb_data = NULL;
    w->options = WFT_VERBOSE | WFT_SKIP_ROOT;
}

int main(int argc, char *argv[])
{
    char *instdir;
    wft_context_t wftctx;

    /* todo: make sure instdir is not under cwd */
    if ((argc < 3)
	|| (! verify_cwd_pkg(argv[1]))
	|| (! dir_exists( (instdir = argv[2]) ))) {
      usage();
      exit(EXIT_FAILURE);
    }

    init_wftctx(&wftctx);

    printf("%s: making regular files read-only (don't edit orig files)\n", PROGNAME);
    if (walk_file_tree(&wftctx) == WFT_ERROR) {
      perror("walk_file_tree");
      exit(EXIT_FAILURE);
    }

    printf("%s: creating directories in %s as needed\n", PROGNAME, instdir);
    wftctx.callback = ensure_dir;
    wftctx.cb_data = instdir;
    if (walk_file_tree(&wftctx) == WFT_ERROR) {
      perror("walk_file_tree");
      exit(EXIT_FAILURE);
    }

    printf("%s: creating hardlinks in %s\n", PROGNAME, instdir);
    wftctx.callback = create_hardlink;
    wftctx.cb_data = instdir;
    if (walk_file_tree(&wftctx) == WFT_ERROR) {
      perror("walk_file_tree");
      exit(EXIT_FAILURE);
    }

    printf("%s: creating symlinks in %s\n", PROGNAME, instdir);
    wftctx.callback = create_symlink;
    wftctx.cb_data = instdir;
    if (walk_file_tree(&wftctx) == WFT_ERROR) {
      perror("walk_file_tree");
      exit(EXIT_FAILURE);
    }

    printf("%s is done.  The following command is recommened:\n  %s\n",
	   PROGNAME, "find . ! -type f -a ! -type d -a ! -type l");

    return 0;
}
