(* oinst.ml - software installation management tool
 * Copyright (C) 2000, 2001, 2002, 2003 Ed L. Cashin
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *)

let ro_fn verbose optdir instdir = 
  function dirname -> function fname ->
    let optpath = optdir ^ "/" ^ dirname ^ "/" ^ fname and
	instpath = instdir ^ "/" ^ dirname ^ "/" ^ fname in
    let perm = (Unix.LargeFile.lstat optpath).Unix.LargeFile.st_perm in
    let new_perm = perm land 0o555 in
    if verbose then Printf.printf "chmod %s 0%o\n" optpath new_perm else ();
    Unix.chmod optpath new_perm;
    Ftw.Continue
   
let umask_val =
  let old = Unix.umask 0 in
  ignore (Unix.umask old);
  old

let mkdir_fn verbose optdir instdir umask = 
   function dirname -> function fname ->
     let optpath = optdir ^ "/" ^ dirname ^ "/" ^ fname in
     if (Unix.LargeFile.lstat optpath).Unix.LargeFile.st_kind != Unix.S_DIR then
       Ftw.Continue
     else
       let instpath = instdir ^ "/" ^ dirname ^ 
	 if fname = "." then
	   "" 
	 else ("/" ^ fname) in
       if Sys.file_exists instpath then
	 ()
       else
	 (let perm = 0o777 land (lnot umask) in
	 if verbose then Printf.printf "mkdir %s 0%o\n" instpath perm else ();
	 Unix.mkdir instpath perm);
       Ftw.Continue

let link_fn verbose optdir instdir = 
  function dirname -> function fname ->
    let optpath = optdir ^ "/" ^ dirname ^ "/" ^ fname in
    let opt_sb = Unix.LargeFile.lstat optpath in
    if opt_sb.Unix.LargeFile.st_kind != Unix.S_REG then
       Ftw.Continue
    else
      let instpath = instdir ^ "/" ^ dirname ^ "/" ^ fname in
      (if verbose then Printf.printf "%s ... " instpath else ();
       try
	let sb = Unix.LargeFile.lstat instpath in
	if opt_sb.Unix.LargeFile.st_ino = sb.Unix.LargeFile.st_ino then
	  (if verbose then print_string "already installed\n" else ())
	else
	  (Bkp.bkp instpath;
	   Unix.link optpath instpath;
	   if verbose then print_string "replaced\n" else ());
      with
	Unix.Unix_error (Unix.ENOENT,_,_) -> 
	  (Unix.link optpath instpath;
	   if verbose then print_string "ok\n" else ());
      );
      Ftw.Continue

let rsync_fn verbose optdir instdir = 
  function dirname -> function fname ->
    let optpath = optdir ^ "/" ^ dirname ^ "/" ^ fname in
    let opt_sb = Unix.LargeFile.lstat optpath in
    let kind = opt_sb.Unix.LargeFile.st_kind in
    if kind != Unix.S_REG && kind != Unix.S_DIR then
      let instpath = instdir ^ "/" ^ dirname ^ "/" ^ fname in
      let cmd = Printf.sprintf "rsync -av %s %s" optpath instpath in
      (if verbose then 
	Printf.printf "Unix.System \"%s\"\n" cmd
      else
	();
       ignore (Unix.system cmd))
    else
      ();
    Ftw.Continue	

let basename s = 
  let slashpos = try 
    (String.rindex s '/') + 1
  with
    Not_found -> 0 in
  String.sub s slashpos ((String.length s) - slashpos)

let usage output =
  Printf.fprintf output
    "usage: %s {pkgdir} {instdir}\n" (basename Sys.argv.(0))

let currdir_ok cdir cwd =
  (basename cwd) = cdir

let main () =
  if (Array.length Sys.argv) < 3 then
    (usage stderr;
     exit 1)
  else
    ();

  let cwd = (Sys.getcwd ()) in

  let currdir = Sys.argv.(1) in
  if currdir_ok currdir cwd then () else (usage stderr; exit 1);

  let instdir = Sys.argv.(2) and 
      verbose = true in
  let ro = ro_fn verbose cwd instdir and
      mkdir = mkdir_fn verbose cwd instdir umask_val and
      link = link_fn verbose cwd instdir and
      rsync = rsync_fn verbose cwd instdir in
  ignore (Ftw.ftw ro cwd);
  ignore (Ftw.ftw mkdir cwd);
  ignore (Ftw.ftw link cwd);
  ignore (Ftw.ftw rsync cwd);
  exit 0;
;;

main()
